/**
 * @file sharqit.h
 * @brief include file of Sharqit library
 */

#ifndef SHARQIT_H
#define SHARQIT_H

#include "fraction.h"
#include "phase.h"
#include "binary_matrix.h"
#include "linear_map.h"
#include "qgate.h"
#include "qcirc.h"
#include "zx.h"
#include "dag.h"
#include "optimizer.h"

namespace Sharqit {
  constexpr char VERSION[] = "v0.2.2"; //!< versoin of sharq
}

#endif
